/*****************************************************************************
 This file is part of the Changedfiles system. 
 $Id: syscalls.c,v 1.13 2003/01/11 13:38:55 rocky Exp $

 This product includes software developed by Konrad Rieck modified
 by Rocky Bernstein.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or (at
 your option) any later version.

 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 USA
******************************************************************************/
/*
  This file contains replacement system calls that live in kernel
  space. These tap off information to pass onto the user area
  changedfiles system.
*/


#define _KERNEL

#include "headers.h"
#include <sys/copyops.h>  /* copyinstr() */
#include <limits.h>       /* PATH_MAX */

extern chf_syscall_t syscalls[];
#ifdef _64BIT
extern chf_syscall_t syscalls32[];
#endif

extern chf_config_t config;
extern krwlock_t config_rwlock;

/* For temporary use. */
static char temp_str[PATH_MAX];
static char temp_str2[PATH_MAX];

/* Things we to to get a read-write lock. */
static void 
set_rwlock(krwlock_t *rwlock) {
    inc_refcnt();
    rw_enter(rwlock, RW_READER);
}

/* Things we to to release the read-write lock we have on. */
static void 
release_rwlock(krwlock_t *rwlock)
{
    rw_exit(&config_rwlock);
    dec_refcnt();
}

/* syslog pid, user and command. */
static void 
syslog_cmd(chf_syscall_op op, const char *old)
{
  struct vnode *vp;
  struct vattr vattr;
  cred_t *cred;

  if (config.debug == OFF) return;

  /* 
     Pick up vnode attributes for file "old" via vp. 
   */
  if (lookupname((char *) old, UIO_USERSPACE, NO_FOLLOW, NULLVPP,
		 &vp))
    goto err;
  
  if (!vp)
    goto err_and_release;
  
  if (VOP_GETATTR(vp, &vattr, 0, CRED()))
    goto err_and_release;
  
  cred = CRED();
  
  dcmn_err((CE_NOTE, "%s(owner %d:%d), process %s (pid %d, uid %d)",
	    op2str[op], vattr.va_uid, vattr.va_gid, u.u_psargs,
	    ddi_get_pid(), cred->cr_ruid));
 err_and_release:
  if (vp)
    VN_RELE(vp);
 err: ;
}


static int64_t 
onefile_op(chf_syscall_op op, const char *file) 
{
  int64_t ret;

  set_rwlock(&config_rwlock);

  syslog_cmd(op, file);
  ret = syscalls[op].sc(file);
  if ( 0 == ret ) {
    /* syscall command did something. Note that. */
    copyinstr (file, temp_str, PATH_MAX, 0);
    dcmn_err((CE_NOTE, "%s(%s) done.", op2str[op], temp_str));
    chf_log_filename(temp_str, NULL, op);
  }
  release_rwlock(&config_rwlock);
  return ret;
}

#ifdef FIXED
/* For reasons I don't understand the rc below comes out 14 rather than
   whatever it really is. The syscall however seems to work.
*/
static int64_t 
twofile_op(const char *old, const char *new, chf_syscall_op op) 
{
  int64_t ret;

  set_rwlock(&config_rwlock);

  syslog_cmd(op, op_str, old);
  ret = syscalls[op].sc(old);
  dcmn_err((CE_NOTE, "ret1 is %d", ret));
  if ( 0 == ret ) {
    /* syscall command did something. Note that. */
    copyinstr (old, temp_str, PATH_MAX, 0);
    copyinstr (new, temp_str2, PATH_MAX, 0);
    dcmn_err((CE_NOTE, "%s(%s, %s) done.", op_str, temp_str, temp_str2));
    chf_log_filename(temp_str, temp_str2, op);
  }
  dcmn_err((CE_NOTE, "ret is %d", ret));
  release_rwlock(&config_rwlock);
  return ret;
}
#endif

/*
 * creat() redirection.
 */
int64_t 
chf_creat(const char *path, mode_t mode)
{
    int64_t ret;
    chf_syscall_op op=CREAT;

    set_rwlock(&config_rwlock);
    ret = syscalls[op].sc(path, mode);
    if ( 0 == ret ) {
      copyinstr (path, temp_str, PATH_MAX, 0);
      chf_log_filename(temp_str, NULL, op);
      dcmn_err((CE_NOTE, "%s(%s, %o) done.", op2str[op], temp_str, mode));
    }
    release_rwlock(&config_rwlock);
    return ret;
}

/*
 * creat64() redirection.
 */
int64_t 
chf_creat64(const char *path, mode_t mode)
{
    int64_t ret;
    chf_syscall_op op=CREAT64;

    set_rwlock(&config_rwlock);
    ret = syscalls[op].sc(path, mode);
    if ( 0 == ret ) {
      copyinstr (path, temp_str, PATH_MAX, 0);
      chf_log_filename(temp_str, NULL, op);
      dcmn_err((CE_NOTE, "%s(%s, %o) done.", op2str[op], temp_str, mode));
    }
    release_rwlock(&config_rwlock);
    return ret;
}

/*
 * mkdir() redirection.
 */
int64_t 
chf_mkdir(const char *path, mode_t mode)
{
    int64_t ret;
    chf_syscall_op op=MKDIR;

    set_rwlock(&config_rwlock);
    ret = syscalls[op].sc(path, mode);
    if ( 0 == ret ) {
      copyinstr (path, temp_str, PATH_MAX, 0);
      chf_log_filename(temp_str, NULL, op);
      dcmn_err((CE_NOTE, "%s(%s, %o) done.", op2str[op], temp_str, mode));
    }
    release_rwlock(&config_rwlock);
    return ret;
}

/*
 * rmdir() redirection.
 */
int64_t 
chf_rmdir(const char *path)
{
  return onefile_op(RMDIR, path);
}

/*
 * unlink() redirection.
 */
int64_t 
chf_unlink(const char *path)
{
  return onefile_op(UNLINK, path);
}

/*
 * link() redirection.
 */
int64_t 
chf_link(const char *old, const char *new)
{
  int64_t ret;
  chf_syscall_op op=LINK;
  
  set_rwlock(&config_rwlock);
  syslog_cmd(op, old);
  ret = syscalls[op].sc(old, new);
  if ( 0 == ret ) {
    copyinstr (old, temp_str, PATH_MAX, 0);
    copyinstr (new, temp_str2, PATH_MAX, 0);
    chf_log_filename(temp_str, temp_str2, op);
    dcmn_err((CE_NOTE, "%s(%s, %s) done.", op2str[op], temp_str, temp_str2));
  }
  release_rwlock(&config_rwlock);
  return ret;
  /*return twofile_op(LINK, old, new);*/
}

/*
 * symlink() redirection.
 */
int64_t 
chf_symlink(const char *old, const char *new)
{
  int64_t ret;
  chf_syscall_op op=SYMLINK;

  set_rwlock(&config_rwlock);
  syslog_cmd(op, old);
  ret = syscalls[op].sc(old, new);
  if ( 0 == ret ) {
    copyinstr (old, temp_str, PATH_MAX, 0);
    copyinstr (new, temp_str2, PATH_MAX, 0);
    chf_log_filename(temp_str, temp_str2, op);
    dcmn_err((CE_NOTE, "%s(%s, %s) done.", op2str[op], temp_str, temp_str2));
  }
  release_rwlock(&config_rwlock);
  return ret;
  /*return twofile_op(SYMLINK, old, new);*/
}

/*
 * rename() redirection.
 */
int64_t 
chf_rename(const char *old, const char *new)
{
  int64_t ret;
  chf_syscall_op op=RENAME;
  set_rwlock(&config_rwlock);
  syslog_cmd(op, old);
  ret = syscalls[op].sc(old, new);
  if ( 0 == ret ) {
    copyinstr (old, temp_str, PATH_MAX, 0);
    copyinstr (new,      temp_str2, PATH_MAX, 0);
    chf_log_filename(temp_str, temp_str2, op);
    dcmn_err((CE_NOTE, "%s(%s, %s) done.", op2str[op], temp_str, temp_str2));
  }
  release_rwlock(&config_rwlock);
  return ret;
  /*return twofile_op(RENAME, old, new);*/
}

static int64_t 
open_op(chf_syscall_op op, const char *path, int oflag, mode_t mode)
{
  int64_t ret;

  set_rwlock(&config_rwlock);

  syslog_cmd(op, path);
  ret = syscalls[op].sc(path, oflag, mode);
  if ( 0 == ret ) {
    /* syscall command did something. Note that. */
    copyinstr (path, temp_str, PATH_MAX, 0);
    dcmn_err((CE_NOTE, "%s(%s) done.", op2str[op], temp_str));
    chf_log_filename(temp_str, NULL, op);
  }
  release_rwlock(&config_rwlock);
  return ret;
}

/*
 * open64() redirection.
 */
int64_t 
chf_open64(const char *path, int oflag, mode_t mode)
{
  return open_op(OPEN64, path, oflag, mode);
}

/*
 * open() redirection.
 */
int64_t 
chf_open(const char *path, int oflag, mode_t mode)
{
  return open_op(OPEN, path, oflag, mode);
}

#ifdef _64BIT

/*
 * creat() redirection.
 */
int64_t 
chf_creat32(const char *path, mode_t mode)
{
    int64_t ret;
    chf_syscall_op op=CREAT;

    set_rwlock(&config_rwlock);
    ret = syscalls[op].sc(path, mode);
    if ( 0 == ret ) {
      copyinstr (path, temp_str, PATH_MAX, 0);
      chf_log_filename(temp_str, NULL, op);
      dcmn_err((CE_NOTE, "%s(%s, %o) done.", op2str[op], temp_str, mode));
    }
    release_rwlock(&config_rwlock);
    return ret;
}

/*
 * creat64() redirection.
 */
int64_t 
chf_creat32_64(const char *path, mode_t mode)
{
    int64_t ret;
    chf_syscall_op op=CREAT64;

    set_rwlock(&config_rwlock);
    ret = syscalls[op].sc(path, mode);
    if ( 0 == ret ) {
      copyinstr (path, temp_str, PATH_MAX, 0);
      chf_log_filename(temp_str, NULL, op);
      dcmn_err((CE_NOTE, "%s(%s, %o) done.", op2str[op], temp_str, mode));
    }
    release_rwlock(&config_rwlock);
    return ret;
}

static int64_t 
chf_open32_op(chf_syscall_op op, const char *path, int oflag, mode_t mode)
{
  int64_t ret;

  set_rwlock(&config_rwlock);
  syslog_cmd(op, path);
  ret = syscalls32[op].sc(path, oflag, mode);
  if ( 0 == ret ) {
    /* syscall command did something. Note that. */
    copyinstr (path, temp_str, PATH_MAX, 0);
    dcmn_err((CE_NOTE, "%s(%s) done.", op2str[op], temp_str));
    chf_log_filename(temp_str, NULL, op);
  }
  release_rwlock(&config_rwlock);
  return ret;
}


/*
 * open64() redirection.
 */
int64_t 
chf_open64_32(const char *path, int oflag, mode_t mode)
{
  return chf_open32_op(OPEN64, path, oflag, mode);
}

/*
 * open() redirection.
 */
int64_t 
chf_open32(const char *path, int oflag, mode_t mode)
{
  return chf_open32_op(OPEN, path, oflag, mode);
}
#endif				/* _64BIT */
